from blackbelt.apis.hipchat import HipChat


def post_message(message):
    client = HipChat()
    client.post_message(message)


def get_last_errors():
    client = HipChat()

    return client.get_last_errors()



from datetime import datetime, timedelta
import json
import re
from subprocess import check_output
from time import sleep
import webbrowser

import click
import requests

from .config import config
from .handle_trello import (
    get_current_working_ticket,
    get_ticket_ready,
    comment_ticket,
    move_to_deployed
)

from .git import (
    get_current_branch,
    merge as git_merge,
    get_github_repo,
    get_remote_repo_info
)

from .messages import post_message
from .circle import wait_for_tests
from .version import VERSION
from .notify import notify

GITHUB_CLIENT_ID = "c9f51ce9cb320bf86f16"

UA_STRING = "black-belt/%s" % VERSION

PR_PHRASE_PREFIX = "Pull request for"


def get_pr_info(pr_url):
    match = re.match(r".*github.com/(?P<owner>\S+)/{1}(?P<name>\S+)/pull/{1}(?P<number>\d+).*$", pr_url)
    if not match:
        raise ValueError("Cannot parse pull request URL, bad format")
    return match.groupdict()


def get_username():
    url = "https://api.github.com/user"

    headers = {
        'Authorization': "token %s" % config['github']['access_token']
    }

    res = requests.get(url, headers=headers)

    return res.json()['login']


def pull_request(card_url):
    """
    Sends current branch for code review.

    #. Inspects current repository for branches
    #. If CARD_URL parameter is not specified, it inspects ``Doing`` on the :term:`Work Board` for the current working ticket (you should have only one working ticket in ``Doing`` that is assigned only to you). If it is specified, it inspects ``Doing`` on the :term:`Work Board` for the selected working ticket, and will raise an error if ticket is not found.
    #. Creates a pull request that references the trello card and references the PR on the card as well
    #. Moves the card to ``Ready``
    #. Opens the browser with the PR for further editing/review
    """
    branch = get_current_branch()
    repo = get_github_repo()

    if 'github.com' not in repo:
        raise ValueError("Current git origin not on github.com; aborting")

    ticket = get_current_working_ticket(card_url)
    md_link = "[%(name)s](%(url)s)" % ticket

    pr_description = """

%(phrase)s %(link)s.

    """ % {
        'phrase': PR_PHRASE_PREFIX,
        'link': md_link
    }

    repo_info = get_remote_repo_info(repo)

    url = "https://api.github.com/repos/%(owner)s/%(name)s/pulls" % repo_info

    payload = {
        'title': ticket['name'],
        'base': 'master',
        'head': "%(owner)s:%(branch)s" % {'branch': branch, 'owner': repo_info['owner']},
        'body': pr_description
    }

    headers = {
        'Authorization': "token %s" % config['github']['access_token']
    }

    r = requests.post(url, data=json.dumps(payload), headers=headers)

    if r.status_code != 201:
        print r.json()
        raise ValueError("PR ended with status code %s: %s" % (r.status_code, r))

    get_ticket_ready(ticket)

    pr_info = r.json()

    ticket_comment = "Sending [pull request #%(number)s](%(html_url)s)." % pr_info

    comment_ticket(ticket, ticket_comment)

    print "Pull request %(pr_id)s for trello card %(ticket_url)s send!" % {
        'pr_id': pr_info['number'],
        'ticket_url': ticket['url']
    }

    webbrowser.open(pr_info['html_url'])


def verify_merge(pr_info, headers, max_waiting_time=30, retry_time=0.1):
    merge_url = "https://api.github.com/repos/%(owner)s/%(name)s/pulls/%(number)s/merge" % pr_info
    start_time = datetime.now()
    succeeded = False

    def do_request():
        r = requests.get(merge_url, headers=headers)

        if (r.status_code == 404):
            if datetime.now() < start_time + timedelta(seconds=max_waiting_time):
                sleep(retry_time)
                return False
            else:
                raise ValueError("GitHub says PR hasn't been merged yet and I've reached the waiting time of %s seconds" % max_waiting_time)

        elif (r.status_code not in [200, 204]):
            raise ValueError("Can't get PR merge info with status code %s" % r.status_code)

        else:
            return True

    while not succeeded:
        succeeded = do_request()


def merge(pr_url):
    """
    This merges PR on Github into master:

    #. Inspects the current repository and the pull request
    #. Switches to master and brings it up to date
    #. Merges the PR locally and pushes to master
    #. Deletes the merged branch from the remote repository/github

    TODO:

    * Comment the associated Trello card

    """

    pr_info = get_pr_info(pr_url)

    pr_api_url = "https://api.github.com/repos/%(owner)s/%(name)s/pulls/%(number)s" % pr_info

    headers = {
        'Authorization': "token %s" % config['github']['access_token'],
        'User-Agent': UA_STRING
    }

    r = requests.get(pr_api_url, headers=headers)

    if (r.status_code != 200):
        raise ValueError("Cannot retrieve PR info with status code %s: %s" % (r.status_code, r))

    pr = r.json()

    if pr['state'] != 'open':
        raise ValueError("PR is %(state)s instead of still being open; not merging" % pr)

    gh_repo = pr['base']['repo']

    if get_github_repo() not in [
        gh_repo['git_url'],
        gh_repo['ssh_url'],
        gh_repo['clone_url']
    ]:
        raise ValueError("The pull request is for the repository %s, while your origin is set up for %s" % (
            gh_repo['git_url'], get_github_repo()
        ))

    sha = pr['head']['sha']

    merge_sha = git_merge(
        sha=sha,
        message="Merging pull request #%(number)s: %(title)s " % pr
    )

    verify_merge(pr_info, headers)

    # All good, delete branch
    branch_url = "https://api.github.com/repos/%(owner)s/%(name)s/git/refs/heads/%(branch)s" % {
        'owner': pr_info['owner'],
        'name': pr_info['name'],
        'branch': pr['head']['ref']

    }
    r = requests.delete(branch_url, headers=headers)

    if (r.status_code != 204):
        raise ValueError("Failed to delete branch after merging pull request, go do it manually")

    print "#%(number)s merged!" % pr_info
    post_message("@here [%(owner)s/%(name)s] Merged PR #%(number)s: %(title)s (%(commits)s commits, %(comments)s comments)" % {
        'owner': pr_info['owner'],
        'name': pr_info['name'],
        'number': pr_info['number'],
        'title': pr['title'],
        'comments': pr['comments'],
        'commits': pr['commits']
    })

    return {
        'sha': merge_sha,
        'owner': pr_info['owner'],
        'name': pr_info['name'],
        'number': pr_info['number'],
        'description': pr['body'],
        'html_url': pr['html_url'],
        'title': pr['title'],
        'branch': pr['head']['ref']
    }


def get_pr_ticket_id(description):
    match = re.search(PR_PHRASE_PREFIX + ' ' + r"\[.*\]\(https://trello.com/c/(?P<id>\w+)/.*\)", description)
    if not match or not 'id' in match.groupdict():
        raise ValueError("Can't find URL in the PR description")

    return match.groupdict()['id']


def deploy(pr_url):
    """
    Deploys PR to production

    #. Does :ref:`pr-merge`
    #. Inform people on HipChat about the merge and the deployment intent
    #. Prepares Heroku deploy slugs using ``grunt create-slug``
    #. Waits for CircleCI tests to pass
    #. TODO: If they fail, asks for retry
    #. Asks for deploy confirmation
    #. Notify others on HipChat about deploy
    #. Deploys
    #. Creates a release on GitHub, using merged branch name as 'ref'.
    #. If it can figure out related Trello card (looks for "Pull request for <link>"), moves it to "Deployed by" column
    #. Does *not* bring beer yet, unfortunately

    """
    merge_info = merge(pr_url)
    
    repo = get_github_repo()
    repo_info = get_remote_repo_info(repo)

    check_output(['grunt', 'create-slug'])

    print "Waiting for tests to pass..."

    ci_info = wait_for_tests(
        sha=merge_info['sha'],
        owner=merge_info['owner'],
        name=merge_info['name']
    )

    if ci_info['failed']:
        notify('Apiary Deployment', "Tests FAILED for %s" % merge_info['sha'])
        raise ValueError("Circle build failed. TODO: Auto retry.")
    notify('Apiary Deployment', "New version %s ready for deploy" % merge_info['sha'])

    # Insert newline
    print ''

    click.confirm("Ready for deploy! Do you want me to deploy %s as the new version of Apiary?" % merge_info['sha'], abort=True)

    post_message("@here Deploying \"%(title)s\" in 15 seconds" % merge_info)

    sleep(15)

    check_output(['grunt', 'deploy-slug'])

    comment = "Deployed by me with version %s. Please verify it works." % merge_info['sha']

    try:
        ticket_id = get_pr_ticket_id(merge_info['description'])
        move_to_deployed(card_id=ticket_id, comment=comment)
    except ValueError:
        if click.prompt("Moving card failed. Open PR in browser?", default=True):
            webbrowser.open(merge_info['html_url'])

    create_release(ref=merge_info['branch'], payload='', description="Deployed to production", repo_info=repo_info)

def create_release(ref, payload, description, repo_info):
    """ Create release in github after deploy to production """

    url = "https://api.github.com/repos/%(owner)s/%(name)s/deployments" % repo_info

    body = {
        'ref': ref,
        'payload': payload,
        'description': description
    }

    headers = {
        'Authorization': "token %s" % config['github']['access_token']
    }

    r = requests.post(url, data=json.dumps(body), headers=headers)

    if r.status_code != 201:
        print r.json()
        raise ValueError("Create github release ended with status code %s: %s" % (r.status_code, r))





from blackbelt.hipchat import post_message as hipchat_post_message
from blackbelt.slack import post_message as slack_post_message

def post_message(message):
  slack_post_message(message)
  hipchat_post_message(message)



import click

from blackbelt.handle_trello import (
    migrate_label as ml,
    schedule_list as sl,
    next_card as n,
    next_week as nw,
    open_current_working_ticket as cc,
    verify as v
)


@click.group(help='Handle Trello-related actions and integrations.')
def cli():
    pass


@cli.command(name='curcard')
def curcard():
    """ Open current doing card in browser """
    cc()


@cli.command(name='migrate-label')
@click.option('--label', default='', help='Label to migrate away')
@click.option('--board', help='Board to migrate from')
@click.option('--board-to', help='Board to migrate to')
@click.option('--column', help='Column to migrate from')
@click.option('--column-to', help='Column to migrate to')
def migrate_label(*args, **kwargs):
    """
    Usage::

        bb t migrate-label --label="Product: Example" --board="1KsoiV9e" --board-to="lEL8Ch52" --column-to="Prepared buffer"
    """
    ml(**kwargs)


@cli.command(name='schedule-list')
@click.option('--story-list', help='Name of the list that should be converted to cards')
@click.option('--owner', help='Whom to assign a created work card')
@click.option('--label', help='Assign a label to the newly-created card. For now, label must be a color name.')
@click.argument('story_card')
def schedule_list(**kwargs):
    """
    Usage::

        bb t schedule-list [--owner="TrelloUserName" [--story-list="Checklist Name"] [--label="color"] http://trello.com/c/story-card-shortlink

    Takes a TODO checklist on a given Story Card. Converts the items into the Work Overview Cards.

    Story list defaults to "To Do", Owner (that the new work tasks are assigned to) defaults to you.
    """
    sl(**kwargs)


@cli.command()
def next(**kwargs):
    """ Takes the top card from To Do queue, moves it to doing and creates proper branch """
    n(**kwargs)


@cli.command(name='next-week')
def next_week():
    """
    Create new columns on the :term:`Work Board`: `Deployed by <sunday>` and `Verified by <sunday>`::

        bb t next-week
    """
    nw()


@cli.command()
@click.argument('story_card')
def verify(**kwargs):
    """
    Looks through a checklists on :term:`Story`, see whether incomplete items refer to a card and whether the card is in `Deployed by <sunday>` and `Verified by <sunday>` column.

    If so, ask to open them and then to verify them (meaning ticking the checkbox)::

        bb t verify http://trello.com/c/story

    """
    #TODO: Scan through all cards in the "Being worked on" column on story board
    v(**kwargs)




class BlackBeltError(Exception):
    """ Catch me to handle all "expected" error states from this package """

class ConfigurationError(BlackBeltError):
    """ Catch me to handle bad configuration (missing config, expired tokens etc.) """



from slacker import Slacker
from blackbelt.config import config

class Slack(object):
    def __init__(self, token=None):
        if not token:
            token = config['slack']['access_token']
            slack = Slacker(token)
            self.slack = slack
        if not token:
            raise ValueError("Can't do things with Slack without access token. Run bb init.")

        self.token = token

    def post_message(self, message, room='#engine-room'):
        return self.slack.chat.post_message(room, message)


def post_message(message):
    client = Slack()
    client.post_message(message)



import json
from urllib import quote

import requests

from blackbelt.config import config


class HipChat(object):
    endpoint = 'https://api.hipchat.com/v2'

    def __init__(self, token=None):
        if not token:
            token = config['hipchat']['access_token']

        if not token:
            raise ValueError("Can't do things with HipChat without access token. Run bb init.")

        self.token = token

    def get_url(self, url, **kwargs):
        kwargs['token'] = self.token

        url = self.endpoint + url + '?auth_token={token}'
        return url.format(**kwargs)

    def request(self, method, url, data=None):
        if data:
            data = json.dumps(data)
        res = getattr(requests, method)(url, data=data, headers={
            'Accept': 'application/json',
            'Content-Type': 'application/json'
        })
        if not res.ok:
            raise ValueError("Request to URL %s failed" % url)

        if res.text:
            return res.json()

    def post_message(self, message, room='Engine Room'):
        url = self.get_url(
            '/room/{room}/notification',
            room=quote(room)
        )

        return self.request('post', url, data={
            'color': 'purple',
            'message': message,
            'message_format': 'text'
        })

    def get_last_errors(self, room='Engine Room'):
        url = self.get_url(
            '/room/{room}/history/latest',
            room=quote(room)
        )

        messages = self.request('get', url)

        papertrail_messages = [m for m in messages['items'] if m['from'] == 'Papertrail']

        return papertrail_messages



import click
import os

plugin_folder = os.path.join(os.path.dirname(__file__), 'commands')


class BlackBelt(click.MultiCommand):

    def list_commands(self, ctx):
        rv = []
        for filename in os.listdir(plugin_folder):
            if filename.endswith('.py') and filename != '__init__.py':
                rv.append(filename[:-3])
        rv.sort()
        return rv

    def get_command(self, ctx, name):
        ns = {}
        fn = os.path.join(plugin_folder, name + '.py')
        with open(fn) as f:
            code = compile(f.read(), fn, 'exec')
            eval(code, ns, ns)
        return ns['cli']

cli = BlackBelt(help='Black Belt: automate project The Apiary Way. Please provide a command.')


# backward compatibility
def main():
    cli()

if __name__ == '__main__':
    cli()



from nose.tools import assert_equals, assert_raises

from blackbelt.git import get_remote_repo_info


class TestGithubRepoGitAddressParsing(object):

    github_repo = 'git@github.com:apiaryio/apiary-test.git'

    def setUp(self):
        self.parsed = get_remote_repo_info(self.github_repo)

    def test_repo_owner(self):
        assert_equals('apiaryio', self.parsed['owner'])

    def test_repo_name(self):
        assert_equals('apiary-test', self.parsed['name'])


class TestGithubRepoHttpsAddressParsing(object):

    github_repo = 'https://github.com/apiaryio/apiary-test.git'

    def setUp(self):
        self.parsed = get_remote_repo_info(self.github_repo)

    def test_repo_owner(self):
        assert_equals('apiaryio', self.parsed['owner'])

    def test_repo_name(self):
        assert_equals('apiary-test', self.parsed['name'])



import mock
from mock import patch
from nose.tools import assert_equals

import datetime

import blackbelt
from blackbelt.handle_trello import infer_branch_name, get_next_sunday

datetime_patcher = mock.patch.object(
    blackbelt.handle_trello.datetime, 'date',
    mock.Mock(wraps=datetime.datetime)
)


class TestInferringBranch(object):

    def test_simple_url(self):
        assert_equals('bb-t-next', infer_branch_name(
            'https://trello.com/c/7b4Z3V8o/1803-bb-t-next'
        ))


class TestNextSunday(object):

    def setUp(self):
        self.mocked_datetime = datetime_patcher.start()

    def tearDown(self):
        datetime_patcher.stop()

    def test_middle_week(self):
        self.mocked_datetime.today.return_value = datetime.datetime(2015, 4, 1) # April's Fools! In fact, Wednesday

        sun = get_next_sunday()
        assert_equals(sun, datetime.date(2015, 4, 5))

    def test_monday(self):
        self.mocked_datetime.today.return_value = datetime.datetime(2015, 4, 6)

        sun = get_next_sunday()
        assert_equals(sun, datetime.date(2015, 4, 12))

    def test_sunday_yields_next_sunday(self):
        self.mocked_datetime.today.return_value = datetime.datetime(2015, 4, 5)

        sun = get_next_sunday()
        assert_equals(sun, datetime.date(2015, 4, 12))



import click

from blackbelt.hipchat import (
    post_message,
    get_last_errors,
)


@click.group(help='Handle HipChat-related tasks and integrations.')
def cli():
    pass


@cli.command()
@click.argument('message', required=True)
def post(message):
    post_message(message)


@cli.command()
def last_errors():
    print get_last_errors()



import click

from blackbelt.slack import (
    post_message
)


@click.group(help='Handle Slack-related tasks and integrations.')
def cli():
    pass


@cli.command()
@click.argument('message', required=True)
def post(message):
    post_message(message)



import datetime
import re
from subprocess import check_output
import webbrowser

import click

#from blackbelt.apis.trello import *
from blackbelt.config import config

from .apis.trello import Trello as TrelloApi

__all__ = ("schedule_list", "migrate_label", "schedule_list")


STORY_CARD_TODO_LIST_NAMES = [
    "To Do",
    "ToDo",
    "Engineering ToDo"
]

TODO_QUEUE_NAME = "To Do Queue"
DEPLOY_QUEUE_NAME = "Ready"
DEPLOYED_PREFIX = "Deployed by"
VERIFIED_PREFIX = "Verified by"


def get_token_url():
    return TrelloApi().get_token_url("black-belt")


def get_api():
    return TrelloApi()


def get_column(name, board_id=None):
    api = get_api()

    if not board_id:
        board_id = config['trello']['work_board_id']

    columns = api.get_columns(board_id)
    column = None

    for col in columns:
        if col['name'] == name:
            column = col

    if not column:
        raise ValueError("Cannot find column %s" % name)

    return column


def get_next_todo_card():
    api = get_api()

    column = get_column(name=TODO_QUEUE_NAME)
    cards = api.get_cards(column_id=column['id'])

    me = api.get_myself()

    my_cards = [card for card in cards if me['id'] in card['idMembers']]

    if len(my_cards) < 1:
        raise ValueError("No card assigned to you in the To Do Queue -- be happy, your job is done!")

    return my_cards[0]


def get_current_working_ticket(card_url=None):
    api = get_api()

    column = get_column(name=config['trello']['work_column_name'])

    cards = api.get_cards(column_id=column['id'])

    me = api.get_myself()

    my_cards = [card for card in cards if me['id'] in card['idMembers']]
    work_card = None

    if len(my_cards) < 1:
        raise ValueError("No working card in your DOING list; aborting.")

    if not card_url:

        if len(my_cards) == 1:
            work_card = my_cards[0]

        if len(my_cards) > 1:
            for card in my_cards:
                if len(card['idMembers']) == 1:
                    if not work_card:
                        work_card = card
                    else:
                        raise ValueError("Multiple available work cards; Try bb gh pr <card_url> ")
    else:
        url_cards = [card for card in my_cards if card['url'] == card_url]

        if len(url_cards) > 0:
            work_card = url_cards[0]

    if not work_card:
        raise ValueError("The specified card_url is not in your DOING list")

    return work_card


def open_current_working_ticket():
    ticket = get_current_working_ticket()
    webbrowser.open(ticket['url'])


def get_ticket_ready(ticket):
    api = get_api()
    column = get_column(name=DEPLOY_QUEUE_NAME)
    api.move_card(card_id=ticket['id'], column_id=column['id'])


def comment_ticket(ticket, comment):
    api = get_api()
    api.comment_card(card_id=ticket['id'], comment=comment)


def migrate_label(label, board, board_to, column, column_to):
    """
    All cards with the given label is moved from :term:`Work Board` into given column in the :term:`Product Board`. ::
    """
    api = get_api()

    if column:
        raise ValueError("column is now ignored, you need to program support for it")

    board_info = api.get_board(board_id=board)

    final_label = None

    if label in board_info['labelNames'].keys():
        final_label = label
    else:
        for l in board_info['labelNames']:
            if board_info['labelNames'][l] == label:
                final_label = l

    if not final_label:
        raise ValueError("Cannot find label %s on given board")

    cards = api.get_cards(board_id=board)

    filtered_cards = []

    for c in cards:
        for l in c['labels']:
            if l['color'] == final_label:
                filtered_cards.append(c)

    board_to_info = api.get_board(board_id=board_to)
    board_to_columns = api.get_columns(board_id=board_to_info['id'])

    for c in board_to_columns:
        if c['name'] == column_to:
            target_column = c

    if not target_column:
        raise ValueError("Cannot find target column %s" % column_to)

    for card in filtered_cards:
        print("Moving card %(id)s: %(name)s" % card)
        api.move_card(
            card_id=card['id'],
            board_id=target_column['idBoard'],
            list_id=target_column['id']
        )


def get_conversion_items(api, card_list, story_card, story_list):
    """ Return (todo_list, items_to_convert_from_there) """
    todo_list = None

    for item in card_list:
        if story_list:
            if item['name'] == story_list or item['id'] == story_list:
                todo_list = item
                break
        else:
            for name in STORY_CARD_TODO_LIST_NAMES:
                if item['name'].lower().strip() == name.strip().lower():
                    todo_list = item
                    break

    if not todo_list:
        lists = ', '.join([i['name'] for i in card_list])
        raise ValueError("Cannot find checklist to convert. Please provide a correct --story-list parameter. Available lists are: %s" % lists)

    list_items = api.get_checklist_items(checklist_id=todo_list['id'])
    return (todo_list, [c for c in list_items if c['state'] == 'incomplete' and not c['name'].startswith('https://trello.com/c/')])


def schedule_list(story_card, story_list=None, owner=None, label=None):
    """
    Looks for Story Card, finds a list and migrate all non-card items to card,
    replacing the items with links to them.

    Work cards contain "Part of <parent-card-link>".
    """

    api = get_api()

    match = re.match("^https\:\/\/trello\.com\/c\/(?P<id>\w+)$", story_card)
    if match:
        story_card = match.groupdict()['id']

    story_card = api.get_card(card_url=story_card)
    #FIXME: list vs. lists
    card_list = api.get_card_checklists(card_id=story_card['id'])

    if owner:
        owner = api.get_member(member_name=owner)

    work_queue = get_column(TODO_QUEUE_NAME)

    todo_list, conversion_items = get_conversion_items(api, card_list, story_card, story_list)

    for item in conversion_items:
        desc = "Part of %(url)s" % story_card
        card = api.create_card(
            name=item['name'],
            description=desc,
            list_id=work_queue['id']
        )

        api.create_item(checklist_id=todo_list['id'], name=card['url'], pos=item['pos'])

        api.delete_checklist_item(checklist_id=todo_list['id'], checklist_item_id=item['id'])

        if owner:
            api.add_card_member(card_id=card['id'], member_id=owner['id'])

        if label:
            api.label_card(card_id=card['id'], label=label)

    print "Done"


def infer_branch_name(url):
    return '-'.join(url.split('/')[~0].split('-')[1:])


def next_card():
    card = get_next_todo_card()

    ticket_urlname = infer_branch_name(card['url'])

    # set up local branch
    from .handle_github import get_username, get_current_branch

    prefix = get_username().lower()

    branch_name = prefix + '/' + ticket_urlname

    if get_current_branch() != 'master':
        check_output(['git', 'checkout', 'master'])

    check_output(['git', 'fetch', 'origin'])

    # You think you might do git checkout branch origin/branch?
    # Oh my, silly you. All of those require the origin branch to exists,
    # therefore pushing. We don't want to push at this stage only to trigger
    # a duplicate CI check on what is basically a master.
    # Therefore, time for some dark magic. If it breaks your git, please
    # carry on and sacrifice a chicken to the daemons of Linux/s

    check_output(['git', 'branch', branch_name, 'origin/master'])
    check_output(['git', 'checkout', branch_name])
    check_output(['git', 'config', 'branch."%s".remote' % branch_name, 'origin'])
    check_output(['git', 'config', 'branch."%s".merge' % branch_name, "refs/heads/\"%s\"" % branch_name])
    check_output(['git', 'config', 'branch."%s".rebase' % branch_name, 'true'])

    # move to Doing
    api = get_api()
    api.move_card(
        card_id=card['id'],
        column_id=get_column(name=config['trello']['work_column_name'])['id']
    )

    # open card for review
    webbrowser.open(card['url'])


def move_to_deployed(card_id, comment=None):
    """ If the card is in Ready column, move it to deployed """
    api = get_api()

    column = get_column(name=DEPLOY_QUEUE_NAME)
    card = api.get_card(card_id=card_id)

    if card['idList'] != column['id']:
        print "The card is not in column %(column)s. NOT moving to Deployed for you." % {
            'column': DEPLOY_QUEUE_NAME
        }
    else:

        columns = api.get_columns(board_id=config['trello']['work_board_id'])
        deployed = None

        for col in columns:
            if col['name'].startswith(DEPLOYED_PREFIX):
                deployed = col
                # Use the first "Deployed by" as the old ones are
                # sometimes "trailing" at the end of the board
                break

        if not column:
            print "Can't find \"Deployed by\" column, NOT moving the card for you"
        else:
            api.move_card(card_id=card_id, column_id=deployed['id'])
            if comment:
                api.comment_card(card_id=card_id, comment=comment)


def get_next_sunday():
    diff = datetime.date.today().weekday()
    # On sunday, add week instead
    if diff == 6:
        diff = -1

    return datetime.date.today() + datetime.timedelta(
        days=6 - diff
    )


def next_week():
    sunday = get_next_sunday().isoformat()

    api = get_api()
    api.add_column(
        board_id=config['trello']['work_board_id'],
        name="Deployed by %s" % sunday,
        position=5
    )
    api.add_column(
        board_id=config['trello']['work_board_id'],
        name="Verified by %s" % sunday,
        position=6
    )


def verify(story_card):
    api = get_api()

    story_card = api.get_card(card_url=story_card)
    checklists = api.get_card_checklists(card_id=story_card['id'])

    # We actually don't care how the checklists are structured, only about the card links
    checklist_cards = []
    index = -1
    for l in checklists:
        index += 1
        for list_item in l['checkItems']:
            # We accept the first link in the checklist item
            match = re.match("^.*https://trello.com/c/(?P<id>\w+)/?(.*)", list_item['name'])
            if match:
                if list_item['state'] == u'incomplete':
                    list_item['_black_belt'] = {
                        'card_id': match.groupdict()['id'],
                        'checklist_index': index,  # Not needed now, but preserved for future list updates
                        'checklist_id': l['id']
                    }
                    checklist_cards.append(list_item)

                elif list_item['state'] != u'complete':
                    print "Unknown checklist state %s, skipping" % list_item['state']

    # We have list of cards to check, go through them and discover whether
    # they are Deployed / Verified
    COLUMN_CACHE = {}

    items_done = []
    for item in checklist_cards:
        work_card = api.get_card(card_id=item['_black_belt']['card_id'])
        column_id = work_card['idList']

        if column_id not in COLUMN_CACHE:
            COLUMN_CACHE[column_id] = api.get_column(column_id)

        name = COLUMN_CACHE[column_id]['name']
        if name.startswith(DEPLOYED_PREFIX) or name.startswith(VERIFIED_PREFIX):
            item['_black_belt']['name'] = work_card['name']
            item['_black_belt']['url'] = work_card['url']
            items_done.append(item)
        else:
            pass
            #print "%s still not done" % work_card['name'].encode('utf-8')

    # Go through the done items, ask for review and ask for completing them
    for item in items_done:
        if click.confirm("Do you want to show card %s?" % item['_black_belt']['name'], default=True):
            webbrowser.open(item['_black_belt']['url'])
        if click.confirm("Do you want to check it as verified?"):
            api.check_item(
                checklist_id=item['_black_belt']['checklist_id'],
                item_id=item['id'],
                card_id=story_card['id']
            )



import re
from subprocess import check_output


"""
I wrap functionality for interacting with the local git instance.

I rely on git being available in PATH, and I rely on standard command line
interface.

When reporting bugs, please do be rigorous in saying what shell, pty/tty you've
used and please provide all environment variables as well as locale settings.
"""


def get_github_repo():
    return check_output(['git', 'config', '--get', 'remote.origin.url']).strip()


def get_remote_repo_info(github_repo_info):
    match = re.match(r".*github.com(:|\/)(?P<owner>[a-zA-Z\_\-]+)/{1}(?P<name>[a-zA-Z\-\_]+)\.git$", github_repo_info)
    if not match:
        raise ValueError("Cannot parse repo info. Bad remote?")
    return match.groupdict()


def get_current_branch():
    return check_output(['git', 'rev-parse', '--abbrev-ref', 'HEAD']).strip()


def get_current_sha():
    return check_output(['git', 'rev-parse', 'HEAD']).strip()


def merge(sha, message):
    """ Merge given SHA into the master and return merge commit SHA """

    if get_current_branch() != 'master':
        check_output(['git', 'checkout', 'master'])

    check_output(['git', 'pull'])

    check_output(['git', 'merge', sha, '-m', message])

    check_output(['git', 'push', 'origin', 'master'])

    return get_current_sha()



import json
from os.path import expanduser, exists
import re
import webbrowser

import click

import handle_trello

CONFIG_FILE = expanduser('~/.blackbelt')


def get_token(config, token_url, group_name, address_text="Please generate a token for yourself:"):
    print '*' * 3 + ' ' + group_name + ' ' + '*' * 3
    print address_text + ' ' + token_url

    default_token = None
    if group_name in config and config[group_name]['access_token']:
        default_token = str(config[group_name]['access_token'])

    webbrowser.open(token_url)

    token = click.prompt("Please insert an access token", default_token).strip()

    if group_name not in config:
        config[group_name] = {}

    config[group_name]['access_token'] = token


def get_trello_url(config, key, prompt, id_regexp):
    group_name = 'trello'
    default_value = None

    url_key = key + '_url'

    if group_name in config and url_key in config[group_name] and config[group_name][url_key]:
        default_value = str(config[group_name][url_key])

    value = click.prompt(
        prompt,
        default_value
    )

    match = re.match(id_regexp, value)
    if match:
        value_id = match.groupdict()['id']
    else:
        raise ValueError("Cannot recognize URL; doesn't match %s regexp" % id_regexp)

    if group_name not in config:
        config[group_name] = {}

    config[group_name][url_key] = value
    config[group_name][key + '_id'] = value_id


def configure_blackbelt():
    print("Going to collect all the tokens and store them in ~/.blackbelt")

    config = {
        'trello': {
            'access_token': None
        },
        'github': {
            'access_token': None
        },
        'circleci': {
            'access_token': None
        },
        'hipchat': {
            'access_token': None
        },
        'slack': {
            'access_token': None
        }
    }

    if exists(CONFIG_FILE):
        with open(CONFIG_FILE) as f:
            config = json.loads(f.read())

    get_token(
        group_name='trello',
        config=config,
        token_url=handle_trello.get_token_url()
    )

    board_id_regexp = r"^https://trello\.com/b/(?P<id>\w+)/.*$"

    get_trello_url(
        config=config,
        key='product_board',
        prompt="Please provide an URL to your Product Board",
        id_regexp=board_id_regexp
    )

    get_trello_url(
        config=config,
        key='work_board',
        prompt="Please provide an URL to your Work Overview",
        id_regexp=board_id_regexp
    )

    get_token(
        group_name='github',
        config=config,
        token_url="https://github.com/settings/applications"
    )

    get_token(
        group_name='circleci',
        config=config,
        token_url="https://circleci.com/account/api"
    )

    get_token(
        group_name='hipchat',
        config=config,
        token_url="https://apiary.hipchat.com/account/api"
    )

    get_token(
        group_name='slack',
        config=config,
        token_url="https://api.slack.com/web"
    )

    with open(expanduser('~/.blackbelt'), 'w') as f:
        f.write(json.dumps(config))

    print('All done!')



import click

from blackbelt.handle_github import (
    pull_request,
    merge as do_merge,
    deploy as do_deploy
)


@click.group(help='Handle github-related tasks and integrations.')
def cli():
    pass


@cli.command()
@click.argument('card_url', required=False)
def pr(*args, **kwargs):
    pr_command(*args, **kwargs)


def pr_command(card_url):
    """
    Usage::

        bb gh pr [CARD_URL]

    See :mod:`threading` for more details.
    """
    pull_request(card_url)


@cli.command()
@click.argument('pr_url')
def merge(*args, **kwargs):
    merge_command(*args, **kwargs)


def merge_command(pr_url):
    """
    Usage::

        bb gh merge https://github.com/apiaryio/apiary/pull/1234
    """
    do_merge(pr_url)


@cli.command()
@click.argument('pr_url')
def deploy(*args, **kwargs):
    deploy_command(*args, **kwargs)


def deploy_command(pr_url):
    """
    Usage::

        bb gh deploy https://github.com/apiaryio/apiary/pull/1234
    """
    do_deploy(pr_url)
